/*
 * FILE: $Source: /u/chestnut/ezk/kitot/advanced-os/2/RCS/sh_env.c,v $
 * HEADER: $Id: sh_env.c,v 1.10 1992/02/11 13:56:39 ezk Exp ezk $
 * REVISION HISTORY:
 * $Log: sh_env.c,v $
 * Revision 1.10  1992/02/11  13:56:39  ezk
 * FINAL VERSION (unless I get an extension...)
 *
 * Revision 1.9  1992/02/10  00:30:08  ezk
 * bg and & work now.  bug fixes so that status reported in "jobs" cmd
 * is correct, and also each time you hit newline.
 *
 * Revision 1.8  1992/02/09  00:22:13  ezk
 * one process in fg.  can ^C it or wait until it ends.  shell prompt
 * using PS1 if used.  can suspend a process but not bg/fg it yet.
 *
 * Revision 1.7  1992/02/08  04:30:01  ezk
 * basic pathname lookup done
 *
 * Revision 1.6  1992/02/08  02:02:10  ezk
 * most parsing done.  set/unset/export work.  variable assignment work.
 *
 * Revision 1.5  1992/02/07  05:24:16  ezk
 * unset command fully working
 *
 * Revision 1.4  1992/02/07  04:53:50  ezk
 * built major jobs structures and converted code to use it.
 * basic export command.
 *
 * Revision 1.3  1992/02/06  21:50:46  ezk
 * simple variables (w/o braces) are supported and expanded, even if
 * prefeixed or postfixed by other strings or variables.
 *
 * Revision 1.2  1992/02/06  16:40:07  ezk
 * basic environment structures created and some manipulation of them (but
 * very basic.)
 *
 * Revision 1.1  1992/02/06  10:09:07  ezk
 * Initial revision
 *
 *
 *
 * 
 * Manipulate shell environment.
 * 
 * E6118y - Advanced Operating Systems.
 * Erez Zadok.
 * Columbia U.  Spring 1992.
 */


#include "process.h"

				/* GLOBALS */
char **sh_envp, **unexported;
int shenvp_len, unexp_len, was_bg;

				/* LOCALS */
static void new_var(), replace_var();

static int envcmp();


/*
 * QSORT(3) comapre function that also makes sure to propagate null
 * variables to the bottom of the list.   This closes holes created by
 * unsetting a variable.
 */
static int var_cmp(v, w)
     char **v, **w;
{
  if (*v == NULL_STR  &&  *w != NULL_STR)
    return 1;			/* v > w */
  if (*v != NULL_STR  &&  *w == NULL_STR)
    return -1;			/* v < w */
  if (*v == NULL_STR  &&  *w == NULL_STR)
    return 0;			/* v == w */
  return strcmp(*v, *w);
}

				/* initialize shell environment from envp */
void init_shenvp(envp)
     char **envp;
{ register int i, offset;

  shenvp_len = 0;
  if (envp == NULL_PAGE) {	/* sanity */
    sh_envp = NULL_PAGE;
    return;
  }
				/* count length of environment */
  shenvp_len = 0;
  while (envp[shenvp_len++] != NULL_STR) ;
				/* allocate space for all pointers */
  sh_envp = (char **) calloc(MAX_ARGV, sizeof (char *));
  if (!sh_envp) {
    fprintf(stderr, "no more memory to init_shenvp!\n");
    exit(-1);
  }

/*
 * This offset trick is needed b/c my login shell passes empty strings
 * in envp for some reason (bad coding of unset?!)
 */
  offset = 0;
  for (i=0; i<shenvp_len; ++i) {
    if (envp[i]  &&  envp[i][0] == '\0') {
      offset++;
      continue;			/* skip this empty var */
    }
    new_var(envp[i], &sh_envp[i-offset]);
  }
  shenvp_len -= offset;
				/* sort env */
  qsort((char*) sh_envp, shenvp_len, sizeof(char *), var_cmp);

				/* allocate space unexported env */
  unexported = (char **) calloc(MAX_ARGV, sizeof (char *));
  if (!unexported) {
    fprintf(stderr, "no more memory to init_shenvp!\n");
    exit(-1);
  }
}


/*
 * Add new variable to environment in given position.
 */
static void new_var(v, e)
     char *v, **e;
{
  if (!v) {			/* then null terminating envp */
    *e = NULL_STR;
    return;
  }
  *e = (char *) calloc(strlen(v)+1, sizeof(char));
  if (!*e) {
    fprintf(stderr, "no more memory to new_var!\n");
    exit(-1);
  }
  strcpy(*e, v);
}

/*
 * Unset a variable.
 */
void unset_var(hcp, icp, s)
     proc_t *hcp, *icp;
     char *s;
{ register int i = -1;
				/* first look in exported variable list */
  i = bin_search_env(s, sh_envp, shenvp_len-1);
  if (i >= 0) {
    free(sh_envp[i]);
    sh_envp[i] = NULL_STR;
    if (shenvp_len > 0) {
      qsort((char*) sh_envp, shenvp_len, sizeof(char *), var_cmp);
      --shenvp_len;
    }
  }
				/* not in unexported list */
  i = bin_search_env(s, unexported, unexp_len);
  if (i >= 0) {
    free(unexported[i]);
    unexported[i] = NULL_STR;
    if (shenvp_len > 0) {
      qsort((char*) unexported, unexp_len, sizeof(char *), var_cmp);
      --unexp_len;
    }
  }
  return;			/* else not found.  do nothing */
}


/*
 * Set a variable to a value.  Assume proper var name.
 */
void set_var(s)
     char *s;
{ register int i;
  static char tmp[MAX_STR];	/* static == don't recreate this mem */

  if (!s)			/* sanity */
    return;
				/* find if var is alreay there */
  i = bin_search_env(s, sh_envp, shenvp_len);
  if (i >= 0) {
    replace_var(sh_envp, shenvp_len, i, s);
    return;
  }
				/* maybe in unexported list */
  i = bin_search_env(s, unexported, unexp_len);
  if (i >= 0) {
    replace_var(unexported, unexp_len, i, s);
    return;
  }
			/* Not found at all.  Add var to unexported list. */
  add_var(unexported, &unexp_len, s);
}


/*
 * Replace a variable in an envp ptr with another.  Sort enrties.
 * Assume values are correct.
 */
static void replace_var(envpp, len, pos, str)
  char **envpp;
  int len, pos;			/* len == length of env */
  char *str;
{
  free(envpp[pos]);
  envpp[pos] = NULL_STR;
  envpp[pos] = (char *) calloc(strlen(str)+1, sizeof(char));
  if (envpp[pos] == NULL_STR) {
    fprintf(stderr, "no more memory to replace_var!\n");
    exit(-1);
  }
  strcpy(envpp[pos], str);
  qsort((char*) envpp, len, sizeof(char *), var_cmp);
}

/*
 * Add a new variable to envp ptr.  Sort enrties.
 * Assume values are correct.
 */
void add_var(envpp, lenp, str)
  char **envpp;
  int *lenp;
  char *str;
{
  envpp[*lenp] = (char *) calloc(strlen(str)+1, sizeof(char));
  if (envpp[*lenp] == NULL_STR) {
    fprintf(stderr, "no more memory to add_var!\n");
    exit(-1);
  }
  strcpy(envpp[*lenp], str);
  (*lenp)++;
  qsort((char*) envpp, *lenp, sizeof(char *), var_cmp);
}

/*
 * Del a variable to envp ptr.  Sort enrties.
 * Assume values are correct.
 */
void del_var(envpp, lenp, pos)
     char **envpp;
  int *lenp, pos;
{
  free(envpp[pos]);
  envpp[pos] = NULL_STR;
  qsort((char*) envpp, *lenp, sizeof(char *), var_cmp);
  (*lenp)--; 
}


/*
 * Check if a string conforms to a variable assignment.
 */
int var_set_cmd(s)
     char *s;
{ register int i;
  static char tmp[MAX_STR];	/* static == don't recreate this mem */

  if (!s)
    return FALSE;
  i = strcspn(s, "=");
  if (i >= strlen(s)  ||  i == 0) /* didn't found = sign */
    return FALSE;
  strncpy(tmp, s, i);
  tmp[i] = '\0';		/* null terminate */
  return legal_var_name(tmp);
}

/*
 * Check if a string conforms to a legal variable name.
 */
int legal_var_name(s)
     char *s;
{ register int i;
  for (i=0; i<strlen(s); ++i)
    if (!isalnum(s[i])  &&  s[i] != '_')
      return FALSE;		/* not a legal variable */
  return TRUE;		/* yes. it is a legal var name */
}

     
/*
 * Expand variable to its value (not recursively!)
 */
void expand_variable(var, expn)
     char *var, **expn;
{ char *tmp, *varval;

  if (expn == NULL_PAGE)	/* sanity */
    return;

  tmp = NULL_STR;		/* just to make sure */
  tmp = sh_getenv(var);
  if (!tmp) {			/* var not found in search */
    *expn = NULL_STR;
    return;
  }
				/* search found. alloc and strcpy. */
  *expn = (char *) calloc(strlen(tmp)+1, sizeof(char));
  if (!*expn) {
    fprintf(stderr, "no more memory to expand_variable!\n");
    exit(-1);
  }
				/* copy from after '=' */
  varval = strchr(tmp, '=');
  if (!varval) {		/* big problem if not found here */
    *expn = NULL_STR;		/* b/c it should have already been */
    return;			/* found before, but anyway... */
  }
  strcpy(*expn, varval + 1);	/* skip '=' */
}

/*
 * Return string to value of variable, or NULL_STR if not found.
 */
char *sh_getenv(s)
     char *s;
{ register int i = -1;
				/* first look in exported variable list */
  i = bin_search_env(s, sh_envp, shenvp_len-1);
  if (i >= 0)
    return sh_envp[i];
				/* not in unexported list */
  i = bin_search_env(s, unexported, unexp_len);
  if (i >= 0)
    return unexported[i];
				/* else not found. return NULL */
  return NULL_STR;
}


/*
 * List all [un]exported variables
 */
void print_shenvp(e, len)
     char **e;
     int len;
{ register int i;
  for (i=0; i<len; ++i)
    if (e[i])
      printf("%3d: %s\n", i, e[i]);
}

/*
 * Compare two environment variables.
 */
static int envcmp(e, w)
     char *e, *w;		/* look if w is in e w/ '=' sign */
{ register int le, lw;
  static char tmp[MAX_STR];

  if (!e && !w)			/* null equals null */
    return 0;
  if (e && !w)			/* w null and e not null w>e */
    return -1;			/* b/c NULL is .gt. anything! by def. */
  if (!e && w)			/* w not null and e is null w<e */
    return 1;

  le = strcspn(e, "=");
  if (le < 1) {
    fprintf(stderr, "corrupt env pointer \"%s\"\n", e);
    exit(-1);
  }
  strncpy(tmp, e, le);		/* copy variable's name */
  tmp[le] = '\0';		/* terminate string with \0 */
  if ((lw = strcspn(w, "=")) < 0) /* comparing var to env line */
    return strcmp(tmp, w);	/* may be matching or not */
				/* comparing env line to another! */
  return strncmp(e, w, MAX(lw, le));  
}


/*
 * Binary search for a word in a an array of strings (list) of length size.
 */
int bin_search_env(word, list, size)
     char *word, **list;
     int size;
{ register int max, min, mid, i, rem;
				/* check for some safety conditions */
  if (word == NULL_STR)
    return -2;
  if (list == NULL_PAGE)
    return -3;
  if (size <= 0)
    return -4;

  max = size - 1;
  min = 0;
  do {
    mid = (max + min) / 2;
    i = envcmp(list[mid], word);
    if (i == 0)			/* word found! */
      return mid;
    if (i > 0)
      max = mid;
    else
      min = mid;
  } while (max > min+1);

  if (max == min+1) {
    if ((i = envcmp( list[max], word)) == 0)
      return max;
    if ((i = envcmp( list[min], word)) == 0)
      return min;
  }
  return(-1);			/* if gets here then word was not found */
}

/*
 * Print shell prompt.
 */
void shell_prompt()
{ static char *p;

  if (was_bg) {			/* avoid printing prompt twice */
    was_bg = 0;
    return;
  }
  expand_variable("PS1", &p);
  if (p)
    printf("%s:%s$ ", progname, p);
  else
    printf("%s$ ", progname);
  fflush(stdout);
  free(p);
  p = NULL_STR;
}
