#include <sys/types.h>
#include <sys/time.h>
#include <rpc/rpc.h>
#include <stdio.h>
#include "checksumd.h"


extern int setitimer();
extern int getitimer();
extern int perror();
#define COUNT		001	/* number of times to repeat count */
#define MAX_SEC		600	/* max seconds for timer */
#define START_TIMERS \
	itv_real.it_interval.tv_sec = MAX_SEC; \
	itv_real.it_interval.tv_usec = 0; \
	itv_real.it_value.tv_sec = MAX_SEC; \
	itv_real.it_value.tv_usec = 0; \
	if (setitimer(ITIMER_REAL, &itv_real) < 0) { \
	  perror("virtual itimer"); \
	  exit(1); \
	}

#define END_TIMERS \
	if (getitimer(ITIMER_REAL, &itv_real) < 0) { \
	  perror("virtual itimer"); \
	  exit(1); \
	} \
	time_real = (MAX_SEC - itv_real.it_value.tv_sec) * 1000000 \
	  	- itv_real.it_value.tv_usec;

static long time_real;
static struct itimerval itv_real;



			/* EXTERNALS */
extern int fprintf(), printf(), atoi(), bzero(), callrpc();

int main(argc, argv)
     int argc;
     char **argv;
{
  char *hostname, *pathname;
  int i4p[4], retval;
  register int i;

  if (argc != 3) {
    fprintf(stderr, "Usage: %s hostname pathname\n", argv[0]);
    exit(1);
  }
  hostname = argv[1];

  bzero((char *)&i4p, sizeof(int) * 4);
 
  pathname = argv[2];
  START_TIMERS;
  for (i=0; i<COUNT; ++i)
    retval = callrpc(hostname,
		     CHKSUMD_PROGRAM, CHKSUMD_VERSION, CHKSUMD_schksum,
		     xdr_wrapstring, &pathname,
		     xdr_checksum, &i4p[0]);
  END_TIMERS;
  printf("time is %d microseconds\n", time_real / COUNT);

#if 0
  inode = atoi(argv[2]);
  retval = callrpc(hostname,
		   CHKSUMD_PROGRAM, CHKSUMD_VERSION, CHKSUMD_ichksum,
		   xdr_long, &inode,
		   xdr_checksum, &i4p[0]);
#endif 0

  printf("CALLRPC status=%d\n", retval);
  if (retval == 0) {		/* then reply structure was filled */
    printf("Got checksum: ");
    printf("%08x|%08x|%08x|%08x\n", i4p[0], i4p[1], i4p[2], i4p[3]);
  }

  exit(retval);
}
